<% macro_groups = macros.group_by { |m| m.name } %>
<% unless macro_groups.empty? %>
  <h3>Macros inherited from <%= ancestor.kind %> <code><%= ancestor.link_from(type) %></code></h3>
  <% i = 0 %>
  <% macro_groups.each do |k, v| %>
    <a href="<%= type.path_to(ancestor) %><%= v[0].anchor %>" class="tooltip">
      <span><%= v.map { |m| m.name + m.args_to_html(:highlight) } .join("<br/>") %></span>
    <%= k %></a><%= ", " if i != macro_groups.size - 1 %>
    <% i += 1 %>
  <% end %>
<% end %>
