<% unless other_types.empty? %>
  <h2>
    <%= Crystal::Doc.anchor_link(title) %>
    <%= title %>
  </h2>
  <ul class="other-types-list">
    <% other_types.each_with_index do |other_type, i| %>
      <li class="other-type"><%= other_type.link_from(type) %></li>
    <% end %>
  </ul>
<% end %>
