<!DOCTYPE html>
<html lang="en">
<head>
  <%= HeadTemplate.new(project_info, nil) %>
  <meta name="repository-name" content="<%= project_info.name %>">
  <title><%= project_info.name %> <%= project_info.version %></title>
  <script type="text/javascript">
  CrystalDocs.base_path = "";
  </script>
</head>
<body>

<%= Crystal::Doc::SVG_DEFS %>
<%= Crystal::Doc::SIDEBAR_BUTTON %>
<%= SidebarTemplate.new(project_info, types, nil) %>

<div class="main-content">
<%= body %>
</div>
</body>
</html>
