lib LibC
  E2BIG           =  1_u16
  EACCES          =  2_u16
  EADDRINUSE      =  3_u16
  EADDRNOTAVAIL   =  4_u16
  EAFNOSUPPORT    =  5_u16
  EAGAIN          =  6_u16
  EALREADY        =  7_u16
  EBADF           =  8_u16
  EBADMSG         =  9_u16
  EBUSY           = 10_u16
  ECANCELED       = 11_u16
  ECHILD          = 12_u16
  ECONNABORTED    = 13_u16
  ECONNREFUSED    = 14_u16
  ECONNRESET      = 15_u16
  EDEADLK         = 16_u16
  EDESTADDRREQ    = 17_u16
  EDOM            = 18_u16
  EDQUOT          = 19_u16
  EEXIST          = 20_u16
  EFAULT          = 21_u16
  EFBIG           = 22_u16
  EHOSTUNREACH    = 23_u16
  EIDRM           = 24_u16
  EILSEQ          = 25_u16
  EINPROGRESS     = 26_u16
  EINTR           = 27_u16
  EINVAL          = 28_u16
  EIO             = 29_u16
  EISCONN         = 30_u16
  EISDIR          = 31_u16
  ELOOP           = 32_u16
  EMFILE          = 33_u16
  EMLINK          = 34_u16
  EMSGSIZE        = 35_u16
  EMULTIHOP       = 36_u16
  ENAMETOOLONG    = 37_u16
  ENETDOWN        = 38_u16
  ENETRESET       = 39_u16
  ENETUNREACH     = 40_u16
  ENFILE          = 41_u16
  ENOBUFS         = 42_u16
  ENODEV          = 43_u16
  ENOENT          = 44_u16
  ENOEXEC         = 45_u16
  ENOLCK          = 46_u16
  ENOLINK         = 47_u16
  ENOMEM          = 48_u16
  ENOMSG          = 49_u16
  ENOPROTOOPT     = 50_u16
  ENOSPC          = 51_u16
  ENOSYS          = 52_u16
  ENOTCONN        = 53_u16
  ENOTDIR         = 54_u16
  ENOTEMPTY       = 55_u16
  ENOTRECOVERABLE = 56_u16
  ENOTSOCK        = 57_u16
  ENOTSUP         = 58_u16
  ENOTTY          = 59_u16
  ENXIO           = 60_u16
  EOPNOTSUPP      = LibC::ENOTSUP
  EOVERFLOW       = 61_u16
  EOWNERDEAD      = 62_u16
  EPERM           = 63_u16
  EPIPE           = 64_u16
  EPROTO          = 65_u16
  EPROTONOSUPPORT = 66_u16
  EPROTOTYPE      = 67_u16
  ERANGE          = 68_u16
  EROFS           = 69_u16
  ESPIPE          = 70_u16
  ESRCH           = 71_u16
  ESTALE          = 72_u16
  ETIMEDOUT       = 73_u16
  ETXTBSY         = 74_u16
  EWOULDBLOCK     = LibC::EAGAIN
  EXDEV           = 75_u16
end
