class XML::Node
  enum Type
    NONE               =  0
    ELEMENT_NODE       =  1
    ATTRIBUTE_NODE     =  2
    TEXT_NODE          =  3
    CDATA_SECTION_NODE =  4
    ENTITY_REF_NODE    =  5
    ENTITY_NODE        =  6
    PI_NODE            =  7
    COMMENT_NODE       =  8
    DOCUMENT_NODE      =  9
    DOCUMENT_TYPE_NODE = 10
    DOCUMENT_FRAG_NODE = 11
    NOTATION_NODE      = 12
    HTML_DOCUMENT_NODE = 13
    DTD_NODE           = 14
    ELEMENT_DECL       = 15
    ATTRIBUTE_DECL     = 16
    ENTITY_DECL        = 17
    NAMESPACE_DECL     = 18
    XINCLUDE_START     = 19
    XINCLUDE_END       = 20
    DOCB_DOCUMENT_NODE = 21
  end
end
