# This file exists so you can do:
#
# ```
# require "compiler/crystal/syntax"
# ```
#
# and use the syntax bits of Crystal (lexer, parser, visitor)
# without requiring all the semantic and codegen code.

require "./syntax/*"
