enum XML::AttributeType
  CDATA       =  1
  ID          =  2
  IDREF       =  3
  IDREFS      =  4
  ENTITY      =  5
  ENTITIES    =  6
  NMTOKEN     =  7
  NMTOKENS    =  8
  ENUMERATION =  9
  NOTATION    = 10
end
