require "uri"
require "./http/client"
require "./http/server"
require "./http/log"
require "./http/common"

# The HTTP module contains `HTTP::Client`, `HTTP::Server` and `HTTP::WebSocket` implementations.
#
# NOTE: To use `HTTP`, you must explicitly import it with `require "http"`
module HTTP
end
