.ansi-base16-railscasts-bright .ansi-bright-black-fg, .ansi-base16-railscasts-bright .ansi-black-fg {
    color: #000000;
  }

.ansi-base16-railscasts-bright .ansi-bright-red-fg, .ansi-base16-railscasts-bright .ansi-red-fg {
    color: #fb0120;
  }

.ansi-base16-railscasts-bright .ansi-bright-green-fg, .ansi-base16-railscasts-bright .ansi-green-fg {
    color: #00770A;
  }

.ansi-base16-railscasts-bright .ansi-bright-yellow-fg, .ansi-base16-railscasts-bright .ansi-yellow-fg {
    color: #fda331;
  }

.ansi-base16-railscasts-bright .ansi-bright-blue-fg, .ansi-base16-railscasts-bright .ansi-blue-fg {
    color: #6fb3d2;
  }

.ansi-base16-railscasts-bright .ansi-bright-magenta-fg, .ansi-base16-railscasts-bright .ansi-magenta-fg {
    color: #d381c3;
  }

.ansi-base16-railscasts-bright .ansi-bright-cyan-fg, .ansi-base16-railscasts-bright .ansi-cyan-fg {
    color: #76c7b7;
  }

.ansi-base16-railscasts-bright .ansi-bright-white-fg, .ansi-base16-railscasts-bright .ansi-white-fg {
    color: #ffffff;
  }

.ansi-base16-railscasts-bright .ansi-bright-black-bg, .ansi-base16-railscasts-bright .ansi-black-bg {
    background-color: #000000;
  }

.ansi-base16-railscasts-bright .ansi-bright-red-bg, .ansi-base16-railscasts-bright .ansi-red-bg {
    background-color: #fb0120;
  }

.ansi-base16-railscasts-bright .ansi-bright-green-bg, .ansi-base16-railscasts-bright .ansi-green-bg {
    background-color: #a1c659;
  }

.ansi-base16-railscasts-bright .ansi-bright-yellow-bg, .ansi-base16-railscasts-bright .ansi-yellow-bg {
    background-color: #fda331;
  }

.ansi-base16-railscasts-bright .ansi-bright-blue-bg, .ansi-base16-railscasts-bright .ansi-blue-bg {
    background-color: #6fb3d2;
  }

.ansi-base16-railscasts-bright .ansi-bright-magenta-bg, .ansi-base16-railscasts-bright .ansi-magenta-bg {
    background-color: #d381c3;
  }

.ansi-base16-railscasts-bright .ansi-bright-cyan-bg, .ansi-base16-railscasts-bright .ansi-cyan-bg {
    background-color: #76c7b7;
  }

.ansi-base16-railscasts-bright .ansi-bright-white-bg, .ansi-base16-railscasts-bright .ansi-white-bg {
    background-color: #ffffff;
  }
