<h1>Settings</h1>

<div class="row">
  <div class="input-field col s12">
    <input id="github" name="settingsGithubToken" type="text">
    <label for="github">GitHub Access Token</label>
  </div>
  <div class="col s12">
    Create a <a href="https://github.com/settings/tokens" target="_blank">GitHub Access Token <svg viewBox="0 0 16 16" class="octicon"><use xlink:href="/vendor/octicons-19.5.0/octicons.svg#link-external-16"></use></svg></a> in order to use the <i>share as gist</i> button.
  </div>
</div>

<div class="row">
  <div class="input-field col s12">
    <div class="switch">
      <label>
        Hide types
        <input name="settingsShowTypes" type="checkbox">
        <span class="lever"></span>
        Show types
      </label>
    </div>
  </div>
</div>

<br>
<br>

<div class="row">
  <div class="input-field col s12">
    <input id="runDebounce" name="settingsRunDebounce" type="number">
    <label for="runDebounce">Time to wait before running the code in milliseconds (leave blank for explicit runs only)</label>
  </div>
</div>
