# This file exists so you can do:
#
# ```
# require "compiler/crystal/formatter"
# ```
#
# and use Crystal's formatter programmatically, without
# requiring all the semantic and codegen code.

require "./tools/formatter"
