lib LibC
  FSCTL_SET_REPARSE_POINT = 0x000900A4 # CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 41, METHOD_BUFFERED, FILE_SPECIAL_ACCESS)
  FSCTL_GET_REPARSE_POINT = 0x000900A8 # CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 42, METHOD_BUFFERED, FILE_ANY_ACCESS)
end
