require "lib_c"

lib LibC
  STATUS_FATAL_APP_EXIT          = 0x40000015_u32
  STATUS_DATATYPE_MISALIGNMENT   = 0x80000002_u32
  STATUS_BREAKPOINT              = 0x80000003_u32
  STATUS_ACCESS_VIOLATION        = 0xC0000005_u32
  STATUS_ILLEGAL_INSTRUCTION     = 0xC000001D_u32
  STATUS_FLOAT_DIVIDE_BY_ZERO    = 0xC000008E_u32
  STATUS_FLOAT_INEXACT_RESULT    = 0xC000008F_u32
  STATUS_FLOAT_INVALID_OPERATION = 0xC0000090_u32
  STATUS_FLOAT_OVERFLOW          = 0xC0000091_u32
  STATUS_FLOAT_UNDERFLOW         = 0xC0000093_u32
  STATUS_PRIVILEGED_INSTRUCTION  = 0xC0000096_u32
  STATUS_STACK_OVERFLOW          = 0xC00000FD_u32
  STATUS_CONTROL_C_EXIT          = 0xC000013A_u32
end
