lib LibC
  alias Int8T = SChar
  alias Int16T = Short
  alias Int32T = Int
  alias Int64T = LongLong
  alias UInt8T = UChar
  alias UInt16T = UShort
  alias UInt32T = UInt
  alias UInt64T = ULongLong
end
