lib LibC
  E2BIG         =  7
  EACCES        = 13
  EADDRINUSE    = 48
  EADDRNOTAVAIL = 49
  EAFNOSUPPORT  = 47
  EAGAIN        = 35
  EALREADY      = 37
  EBADF         =  9
  EBADMSG       = 89
  EBUSY         = 16
  ECANCELED     = 85
  ECHILD        = 10
  ECONNABORTED  = 53
  ECONNREFUSED  = 61
  ECONNRESET    = 54
  EDEADLK       = 11
  EDESTADDRREQ  = 39
  EDOM          = 33
  EDQUOT        = 69
  EEXIST        = 17
  EFAULT        = 14
  EFBIG         = 27
  EHOSTUNREACH  = 65
  EIDRM         = 82
  EILSEQ        = 86
  EINPROGRESS   = 36
  EINTR         =  4
  EINVAL        = 22
  EIO           =  5
  EISCONN       = 56
  EISDIR        = 21
  ELOOP         = 62
  EMFILE        = 24
  EMLINK        = 31
  EMSGSIZE      = 40
  EMULTIHOP     = 90
  ENAMETOOLONG  = 63
  ENETDOWN      = 50
  ENETRESET     = 52
  ENETUNREACH   = 51
  ENFILE        = 23
  ENOBUFS       = 55
  ENODEV        = 19
  ENOENT        =  2
  ENOEXEC       =  8
  ENOLCK        = 77
  ENOLINK       = 91
  ENOMEM        = 12
  ENOMSG        = 83
  ENOPROTOOPT   = 42
  ENOSPC        = 28
  ENOSYS        = 78
  ENOTCONN      = 57
  ENOTDIR       = 20
  ENOTEMPTY     = 66
  #  ENOTRECOVERABLE = 95
  ENOTSOCK   = 38
  ENOTSUP    = LibC::EOPNOTSUPP
  ENOTTY     = 25
  ENXIO      =  6
  EOPNOTSUPP = 45
  EOVERFLOW  = 84
  #  EOWNERDEAD      = 96
  EPERM           =  1
  EPIPE           = 32
  EPROTO          = 92
  EPROTONOSUPPORT = 43
  EPROTOTYPE      = 41
  ERANGE          = 34
  EROFS           = 30
  ESPIPE          = 29
  ESRCH           =  3
  ESTALE          = 70
  ETIMEDOUT       = 60
  ETXTBSY         = 26
  EWOULDBLOCK     = LibC::EAGAIN
  EXDEV           = 18
end
