<div class="row">
</div>

<div class="row row-narrow" id="mainEditorContainer">
</div>

<div class="row">
  <div class="col s7">
    <a id="runFormatterBtn" class="code-btn waves-effect waves-light btn-flat" href="#run-formatter">
      <svg viewBox="0 0 16 16" class="octicon"><use xlink:href="/vendor/octicons-19.5.0/octicons.svg#checklist-16"></use></svg>
      <span>Run Formatter</span>
    </a>
  </div>
  <div class="col s5">
    <a class="modal-trigger code-btn waves-effect waves-light btn-flat" href="#output-modal">
      <svg id="mainOutputIndicator" viewBox="0 0 16 16" class="octicon"><use xlink:href="/vendor/octicons-19.5.0/octicons.svg#terminal-16"></use></svg>
      <span>Show Output</span>
    </a>
  </div>
</div>

<div class="fixed-action-btn" style="bottom: 45px; right: 24px;" id="mainButtonsContainer">
  <ul class="unstyled">
    <li>
      <a class="btn-floating btn grey darken-1 tooltipped" href="#" data-position="left" data-delay="50" data-tooltip="Share as gist" id="saveAsGist">
        <svg viewBox="0 0 16 16" class="octicon"><use xlink:href="/vendor/octicons-19.5.0/octicons.svg#github-16"></use></svg>
      </a>
    </li>
    <li>
      <a class="btn-floating btn red darken-1 tooltipped" href="#" data-position="left" data-delay="50" data-tooltip="Download code as file" id="saveAsFile">
        <svg viewBox="0 0 16 16" class="octicon"><use xlink:href="/vendor/octicons-19.5.0/octicons.svg#desktop-download-16"></use></svg>
      </a>
    </li>
  </ul>
</div>

<div id="output-modal" class="modal bottom-sheet">
  <div class="modal-content">
    <h6>Output</h6>
    <pre class="ansi-base16-railscasts-bright"><code class="output" id="mainOutput"></code></pre>
  </div>
</div>
